#!perl

BEGIN {
  unless ($ENV{AUTHOR_TESTING}) {
    print qq{1..0 # SKIP these tests are for testing by the author\n};
    exit
  }
}


use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::Perl::Critic::Subset 3.001.003

use Test::Perl::Critic (-profile => "") x!! -e "";

my $filenames = ['lib/App/CSVUtils.pm','script/csv-add-field','script/csv-avg','script/csv-concat','script/csv-convert-to-hash','script/csv-delete-field','script/csv-dump','script/csv-each-row','script/csv-grep','script/csv-info','script/csv-list-field-names','script/csv-lookup-fields','script/csv-map','script/csv-munge-field','script/csv-replace-newline','script/csv-select-fields','script/csv-select-row','script/csv-setop','script/csv-sort','script/csv-sort-fields','script/csv-sort-rows','script/csv-split','script/csv-sum','script/csv2ltsv','script/csv2tsv','script/dump-csv','script/tsv2csv'];
unless ($filenames && @$filenames) {
    $filenames = -d "blib" ? ["blib"] : ["lib"];
}

all_critic_ok(@$filenames);
